#-------------------------------------------------------------------------------
# Name:       ExtractPolygons
# Purpose:
#
# Author:      jwezorek
#-------------------------------------------------------------------------------
#!/usr/bin/env python

class DirInfoItem:
    def __init__(self, fl, fr, pt):
        self.f_left = fl
        self.f_right = fr
        self.next_poly_point = pt
        self.nil = (0,0)

class StateInfo:
    def __init__(self, e, np, nd):
        self.is_emitted = e
        self.next = np
        self.dir_offset = nd

class ArrayFacade:
    def __init__(self, a):
        self.ary = a
        self.hgt = len(a)
        self.wd = len(a[0])
    def __getitem__(self, xy):
        if (xy[0] < 0 or xy[1] < 0 or xy[0] >= self.wd or xy[1] >= self.hgt):
            return []
        else:
            return self.ary[xy[1]][xy[0]]
    def AllCells(self):
        for y in range(0, self.hgt):
            for x in range(-1, self.wd+1):
                yield (x,y)

def PrintAry(ary):
    for row in ary.ary:
        print row
    print "\n"

def Extract( aryInput, bkgd_color ):

    ary = ArrayFacade(aryInput)
    aryOpenBoundary = ArrayFacade([[[[],[]] for x in range(0,ary.wd+1)] for y in range(0,ary.hgt)])

    polygons = []

    def RightLoc( xy ):
        return (xy[0]+1,xy[1])

    def IsBoundaryLoc( xy ):
        return ary[xy] != ary[RightLoc(xy)]

    class Walker:

        BoundaryCellForDir = {0:1,2:0}
        BoundarySignForDir = {0:1,2:-1}

        def __init__( self, pos, color_offset_x, d ):
            self.dir= d
            self.loc = pos
            self.color = ary[(pos[0]+color_offset_x, pos[1])]

        def MakeTurn(self, offset):
            self.dir += offset
            if (self.dir < 0):
                self.dir = 3
            elif (self.dir > 3):
                self.dir = 0

        def AddLocs(self,l1,l2):
            return (l1[0]+l2[0],l1[1]+l2[1])

        def GetForwardPairLocs(self):
            fl = Walker.DirInfo[self.dir].f_left;
            fr = Walker.DirInfo[self.dir].f_right;
            return (self.AddLocs(fl,self.loc),self.AddLocs(fr,self.loc))

        def GetForwardPairMask(self):
            (fl,fr) = self.GetForwardPairLocs()
            if (ary[fl] == self.color):
                left = 1
            else:
                left = 0
            if (ary[fr] == self.color):
                right = 1
            else:
                right = 0
            return (left,right)

        def GetBoundaryLoc(self):
            if (self.dir in Walker.BoundaryOffsetForDir):
                return (self.loc[0]+Walker.BoundaryOffsetForDir[self.dir],self.loc[1])
            else:
                return []

        def DoOneStep(self, index):
            if (self.GetBoundaryLoc() != []):
                bounds_loc = self.GetBoundaryLoc()
                if (self.dir in Walker.BoundaryOffsetForDir ):
                    aryOpenBoundary[bounds_loc][Walker.BoundaryCellForDir[self.dir]] = index * Walker.BoundarySignForDir[self.dir]
            forw_pair_mask = self.GetForwardPairMask()
            dir_info = Walker.DirInfo[self.dir]
            state_transition = Walker.NextState[forw_pair_mask]

            pt =  self.AddLocs( self.loc, dir_info.next_poly_point)
            next_loc = state_transition.next
            self.loc = self.AddLocs( self.loc, getattr(dir_info, next_loc))
            self.MakeTurn(state_transition.dir_offset)
            if (state_transition.is_emitted):
                return pt
            else:
                return []

        def WalkBoundary(self, index):
            startLoc = self.loc
            startDir = self.dir
            pt = self.DoOneStep(index)
            if (pt != []):
                yield pt
            while(self.loc != startLoc or self.dir != startDir):
                pt = self.DoOneStep(index)
                if (pt != []):
                    yield pt

    class RightWalker(Walker):
        def __init__(self,pos,d):
            Walker.DirInfo = [
                DirInfoItem((0,-1),(1,-1),(1,0)),
                DirInfoItem((1,0),(1,1),(1,1)),
                DirInfoItem((0,1),(-1,1),(0,1)),
                DirInfoItem((-1,0),(-1,-1),(0,0))
            ]
            Walker.NextState = {
                (0,0) : StateInfo(1, 'f_right', 1),
                (0,1) : StateInfo(0, 'f_left', 0),
                (1,0) : StateInfo(1, 'nil', -1),
                (1,1) : StateInfo(1, 'nil', -1)
            }
            Walker.BoundaryOffsetForDir = {0:1,2:0}
            Walker.__init__(self, pos, 1, d)

    class LeftWalker(Walker):
        def __init__(self,pos):
            Walker.DirInfo = [
                DirInfoItem((-1,-1),(0,-1),(0,0)),
                DirInfoItem((1,-1),(1,0),(1,0)),
                DirInfoItem((1,1),(0,1),(1,1)),
                DirInfoItem((-1,1),(-1,0),(0,1))
            ]
            Walker.NextState = {
                (0,0) : StateInfo(1, 'nil', 1),
                (0,1) : StateInfo(0, 'f_right', 0),
                (1,0) : StateInfo(1, 'f_left', -1),
                (1,1) : StateInfo(1, 'f_left', -1)
            }
            Walker.BoundaryOffsetForDir = {0:0,2:1}
            Walker.__init__(self, pos, 0, 2)

    def IsOpenRightBoundary(loc):
        (left_color, right_color) = GetBoundaryColorPair(loc, bkgd_color)
        if (right_color == []):
            return False
        boundary_cells = aryOpenBoundary[RightLoc(loc)]
        return (boundary_cells[1] == [])

    def IsOpenLeftBoundary(loc):
        (left_color, right_color) = GetBoundaryColorPair(loc, bkgd_color)
        if (left_color == []):
            return False
        boundary_cells = aryOpenBoundary[RightLoc(loc)]
        return (boundary_cells[0] == [])

    def GetBoundaryColorPair(loc, bkgd_color):
        left = ary[loc] if ary[loc] != bkgd_color else []
        right = ary[RightLoc(loc)] if ary[RightLoc(loc)] != bkgd_color else []
        return (left,right)

    output = []
    curr_poly_index = []
    for p in ary.AllCells():
        (left_color, right_color) = GetBoundaryColorPair(p, bkgd_color)
        if (left_color != right_color):

            if (IsOpenLeftBoundary(p)):
                if (curr_poly_index == []):
                    raise  NameError('HoleNotInPolygon')
                walker = LeftWalker(p)
                new_poly = []
                for pt in walker.WalkBoundary(curr_poly_index):
                    new_poly.append(pt)
                polygons[curr_poly_index-1][0].append(new_poly)
                #PrintAry(aryOpenBoundary)
            if (IsOpenRightBoundary(p)):
                new_poly_index = len(polygons)+1
                walker = RightWalker(p,0)
                new_poly = []
                for pt in walker.WalkBoundary(new_poly_index):
                    new_poly.append(pt)
                polygons.append([[new_poly], right_color])
                #PrintAry(aryOpenBoundary)
            boundary_indices = aryOpenBoundary[RightLoc(p)]
            curr_poly_index = boundary_indices[1] if boundary_indices[1] > 0 else []

    return polygons






